#include "matrix.hpp"

Matrix::Matrix(usint _p,usint _q){
  p=_p;
  q=_q;
  coeffs=new Rational[p*q];
}

Matrix::Matrix(const Matrix& M){
  p=M.p;
  q=M.q;
  coeffs=new Rational[p*q];
  memcpy(coeffs,M.coeffs,p*q*sizeof(Rational));
}

Matrix::~Matrix(){
  delete[] coeffs;
}

Matrix& Matrix::operator=(const Matrix& M){
  if(p*q!=M.p*M.q){
    delete[] coeffs;
    coeffs=new Rational[M.p*M.q];
  }
  p=M.p;
  q=M.q;
  memcpy(coeffs,M.coeffs,p*q*sizeof(Rational));
  return *this;
}

void Matrix::set(initializer_list<Rational> list){
  assert(list.size()==usint(p*q));
  usint index=0;
  for(initializer_list<Rational>::iterator it=list.begin();it!=list.end();++it){
    coeffs[index++]=*it;
  }
}

usint Matrix::lines() const{
  return p;
}

usint Matrix::columns() const{
  return q;
}

void Matrix::set(usint i,usint j,const Rational& x){
  assert(1<=i and i<=p and 1<=j and j<=q);
  coeffs[j-1+(i-1)*q]=x;
}

Rational Matrix::get(usint i,usint j) const{
  assert(1<=i and i<=p and 1<=j and j<=q);
  return coeffs[j-1+(i-1)*q];
}

Rational& Matrix::getRef(usint i,usint j){
  assert(1<=i and i<=p and 1<=j and j<=q);
  return coeffs[j-1+(i-1)*q];
}

bool Matrix::operator==(const Matrix& M) const{
  if(p!=M.p) return false;
  if(q!=M.q) return false;
  for(usint i=0;i<p*q;++i){
    if(coeffs[i]!=M.coeffs[i]) return false;
  }
  return true;
}

bool Matrix::operator!=(const Matrix& M) const{
  if(p!=M.p) return true;
  if(q!=M.q) return true;
  for(usint i=0;i<p*q;++i){
    if(coeffs[i]!=M.coeffs[i]) return true;
  }
  return false;
}

Matrix Matrix::operator+(const Matrix& M) const{
  assert(p==M.p and q==M.q);
  Matrix res(p,q);
  for(usint i=0;i<p*q;++i){
    res.coeffs[i]=coeffs[i]+M.coeffs[i];
  }
  return res;
}

Matrix Matrix::operator-(const Matrix& M) const{
  assert(p==M.p and q==M.q);
  Matrix res(p,q);
  for(usint i=0;i<p*q;++i){
    res.coeffs[i]=coeffs[i]-M.coeffs[i];
  }
  return res;
}

Matrix Matrix::operator*(const Rational& r) const{
  Matrix res(p,q);
  for(usint i=0;i<p*q;++i){
    res.coeffs[i]=r*coeffs[i];
  }
  return res;
}

Matrix operator*(const Rational& r,const Matrix& M){
  Matrix res(M.p,M.q);
  for(usint i=0;i<M.p*M.q;++i){
    res.coeffs[i]=r*M.coeffs[i];
  }
  return res;
}

Matrix Matrix::operator-() const{
  Matrix res(p,q);
  for(usint i=0;i<p*q;++i){
    res.coeffs[i]=-coeffs[i];
  }
  return res;
}

Matrix Matrix::operator*(const Matrix& A) const{
  assert(q==A.p);
  Matrix res(p,A.q);
  for(int i=1;i<=p;++i){
    for(int j=1;j<=A.q;++j){
      Rational s(0);
      for(int k=1;k<=q;++k){
	s=s+get(i,k)*A.get(k,j);
      }
      res.set(i,j,s);
    }
  }
  return res;
}

void Matrix::swapLines(usint i,usint j){
  Matrix& A=*this;//A is the current matrix
  Rational temp;
  for(usint k=1;k<=q;++k){
    temp=A[i][k];
    A[i][k]=A[j][k];
    A[j][k]=temp;
  }
}

void Matrix::mulLine(usint i,Rational c){
  Matrix& A=*this;//A is the current matrix
  for(usint j=1;j<=q;++j){
    A[i][j]=c*A[i][j];
  }
}

void Matrix::addMulLine(usint i,usint j, Rational c){
  Matrix& A=*this;//A is the current matrix
  for(usint k=1;k<=q;++k){
    A[i][k]=A[i][k]+c*A[j][k];
  }
}


//Recherche d'une pivot sur la colonne C_i à partir de la position j
//Retourne la ligne du pivot (s'il existe) et 0 sinon
int Matrix::findPivot(usint i,usint j){
  for(int k=j;k<=p;++k){
    if(get(k,i)!=0) return k;
  }
  return 0;
} 

void Matrix::Gauss(){
  Matrix& A=*this;
  //cp : colonne pivot
  //lp : ligne pivot
  int lp=1;
  for(int cp=1;cp<=q;++cp){
    int temp=A.findPivot(cp,lp);
    if(temp!=0){
      A.swapLines(lp,temp);
      A.mulLine(lp,1/A[lp][cp]);
      for(int j=1;j<=p;++j){
	if(j!=lp){
	  A.addMulLine(j,lp,-A[j][cp]);
	}
      }
      ++lp;
    }
  }
}

Rational Matrix::determinant() const{
  assert(p==q);
  Matrix A(*this); //A est une copie
  Rational det(1);
  int lp=1;
  for(int cp=1;cp<=A.q;++cp){
    int temp=A.findPivot(cp,lp);
    if(temp!=0){
      det=-det;
      A.swapLines(lp,temp);
      Rational r=A[lp][cp];
      det=r*det;
      A.mulLine(lp,1/r);
      for(int j=1;j<=A.p;++j){
	if(j!=lp){
	  A.addMulLine(j,lp,-A[j][cp]);
	}
      }
      ++lp;
    }
    else{
      return 0;
    }
  }
  return det;
}

bool Matrix::isInvertible() const{
  if(p!=q) return false;
  return determinant()!=0;
}


Matrix Matrix::inverse() const{
  if(not isInvertible()) throw("The matrix is not invertible");
  int n=p;
  Matrix M(n,2*n);
  const Matrix& A=*this;
  for(int i=1;i<=n;++i){
    for(int j=1;j<=n;++j){
      M[i][j]=A[i][j];
    }
    M[i][n+i]=1;
  }
  M.Gauss();
  Matrix res(n,n);
  for(int i=1;i<=n;++i){
    for(int j=1;j<=n;++j){
      res[i][j]=M[i][n+j];
    }
  }
  return res;
}

int Matrix::rank() const{
  Matrix A(*this); // copie de la matrice courante, nommée A
  A.Gauss();
  int j=1;
  for(int i=1;i<=A.p;++i){
    while(j<=A.q and A[i][j]==0) ++j;
    if(j>A.q) return i-1;
  }
  return A.p;
}


ostream& operator<<(ostream& os,const Matrix& M){
  usint offset=0;
  os<<endl;
  for(usint i=0;i<M.p;++i){
    os<<'[';
    for(usint j=0;j<M.q;++j){
      os<<'\t'<<M.coeffs[offset];
      ++offset;
    }
    os<<"]"<<endl;
  }
  return os;
}


